@@ -91,14 +91,14 @@ freebsd_install() {
     libtool
     gettext-tools
     python
-    ruby26
+    lang/ruby27
   )
 
   # Note: we assume sudo is already installed
   "${SUDO}" pkg install -y "${packages[@]}"
 
   cd /usr/ports/devel/ruby-gems
-  "${SUDO}" make -DBATCH RUBY_VER=2.5 install
+  "${SUDO}" make -DBATCH RUBY_VER=2.7 install
   cd
 
   mkdir -p ~/.gnupg
@@ -118,6 +118,7 @@ netbsd_install() {
 
 linux_prepare_ribose_yum_repo() {
   "${SUDO}" rpm --import https://github.com/riboseinc/yum/raw/master/ribose-packages.pub
+  "${SUDO}" rpm --import https://github.com/riboseinc/yum/raw/master/ribose-packages-next.pub
   "${SUDO}" curl -L https://github.com/riboseinc/yum/raw/master/ribose.repo \
     -o /etc/yum.repos.d/ribose.repo
 }
@@ -152,8 +153,11 @@ linux_install_centos() {
     centos-8)
       linux_install_centos8
       ;;
+    centos-9)
+      linux_install_centos9
+      ;;
     *)
-      >&2 echo "Error: unsupported CentOS version \"${DIST_VERSION_ID}\" (supported: 7, 8).  Aborting."
+      >&2 echo "Error: unsupported CentOS version \"${DIST_VERSION_ID}\" (supported: 7, 8, 9).  Aborting."
       exit 1
   esac
 }
@@ -248,9 +252,25 @@ linux_install_centos7() {
 }
 
 linux_install_centos8() {
+  "${SUDO}" "${YUM}" config-manager --set-enabled powertools
   yum_prepare_repos epel-release
   yum_install_build_dependencies \
-    cmake
+    cmake \
+    texinfo
+
+  yum_install_dynamic_build_dependencies_if_needed
+
+  ensure_automake
+  ensure_cmake
+  ensure_ruby
+  rubygem_install_build_dependencies
+}
+
+linux_install_centos9() {
+  yum_prepare_repos epel-release
+  yum_install_build_dependencies \
+    cmake \
+    texinfo
 
   yum_install_dynamic_build_dependencies_if_needed
 
@@ -292,7 +312,10 @@ install_static_cacheable_build_dependencies() {
 
   mkdir -p "$LOCAL_BUILDS"
 
-  local default=(botan jsonc gpg)
+  local default=(jsonc gpg)
+  if [ "${CRYPTO_BACKEND:-}" != "openssl" ]; then
+    default=(botan "${default[@]}")
+  fi
   local items=("${@:-${default[@]}}")
   for item in "${items[@]}"; do
     install_"$item"
@@ -609,6 +632,7 @@ msys_install() {
     gnupg2
     make
     pkg-config
+    p7zip
     mingw64/mingw-w64-x86_64-cmake
     mingw64/mingw-w64-x86_64-python3
   )
